"""
Library to handle Dynet networks.

@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import json

from .opcodes import OpcodeType, SyncType


class PacketError(Exception):
    """Class for Dynet packet errors."""

    def __init__(self, message):
        """Initialize the error."""
        self.message = message
        super().__init__(message)


class DynetPacket:
    """Class for a Dynet network packet."""

    def __init__(self, msg=None):
        """Initialize the packet."""
        self.opcode_type = None
        self.sync = None
        self.area = None
        self.data = []
        self.command = None
        self.join = None
        self.chk = None
        if msg is not None:
            self.from_msg(msg)

    def to_msg(self, area, command, data, sync=SyncType.LOGICAL.value, join=255):
        """Convert packet to a binary message."""
        my_bytes = []
        my_bytes.append(sync)
        my_bytes.append(area)
        my_bytes.append(data[0])
        my_bytes.append(command)
        my_bytes.append(data[1])
        my_bytes.append(data[2])
        my_bytes.append(join)
        my_bytes.append(self.calc_sum(my_bytes))
        self.from_msg(my_bytes)

    def from_msg(self, msg):
        """Decode a Dynet message."""
        message_length = len(msg)
        if message_length < 8:
            raise PacketError(f"Message too short ({len(msg)} bytes): {msg}")
        if message_length > 8:
            raise PacketError(f"Message too long ({len(msg)} bytes): {msg}")
        self._msg = msg
        self.sync = self._msg[0]
        self.area = self._msg[1]
        self.data = [self._msg[2], self._msg[4], self._msg[5]]
        self.command = self._msg[3]
        self.join = self._msg[6]
        self.chk = self._msg[7]
        if self.sync == 28:
            if OpcodeType.has_value(self.command):
                self.opcode_type = OpcodeType(self.command).name

    @staticmethod
    def calc_sum(msg):
        """Calculate the checksum."""
        msg = msg[:7]
        return -(sum(ord(c) for c in "".join(map(chr, msg))) % 256) & 0xFF

    def __repr__(self):
        """Print the packet."""
        return json.dumps(self.__dict__)

    @staticmethod
    def set_channel_level_packet(area, channel, level, fade):
        """Create a packet to set level of a channel."""
        channel_bank = 0xFF if (channel <= 4) else (int((channel - 1) / 4) - 1)
        target_level = int(255 - 254 * level)
        opcode = 0x80 + ((channel - 1) % 4)
        fade_time = int(fade / 0.02)
        if (fade_time) > 0xFF:
            fade_time = 0xFF
        packet = DynetPacket()
        packet.to_msg(
            area=area,
            command=opcode,
            data=[target_level, channel_bank, fade_time],
        )
        return packet

    @staticmethod
    def select_area_preset_packet(area, preset, fade):
        """Create a packet to select a preset in an area."""
        preset = preset - 1
        bank = int((preset) / 8)
        opcode = preset - (bank * 8)
        if opcode > 3:
            opcode = opcode + 6
        fade_low = int(fade / 0.02) - (int((fade / 0.02) / 256) * 256)
        fade_high = int((fade / 0.02) / 256)
        packet = DynetPacket()
        packet.to_msg(
            area=area, command=opcode, data=[fade_low, fade_high, bank],
        )
        return packet

    @staticmethod
    def request_channel_level_packet(area, channel):
        """Create a packet to request the level of a specific channel."""
        packet = DynetPacket()
        packet.to_msg(
            area=area,
            command=OpcodeType.REQUEST_CHANNEL_LEVEL.value,
            data=[channel - 1, 0, 0],
        )
        return packet

    @staticmethod
    def stop_channel_fade_packet(area, channel):
        """Create a packet to stop fade of a channel."""
        packet = DynetPacket()
        packet.to_msg(
            area=area,
            command=OpcodeType.STOP_FADING.value,
            data=[channel - 1, 0, 0],
        )
        return packet

    @staticmethod
    def request_area_preset_packet(area):
        """Create a packet to request the current preset in an area."""
        packet = DynetPacket()
        packet.to_msg(
            area=area,
            command=OpcodeType.REQUEST_PRESET.value,
            data=[0, 0, 0],
        )
        return packet



"""
Manage a Dynalite connection.

@ Author      : Troy Kelly
@ Date        : 3 Dec 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import asyncio
import time

from .const import (
    CONF_ACTION,
    CONF_ACTION_CMD,
    CONF_AREA,
    CONF_CHANNEL,
    CONF_PRESET,
    CONF_TRGT_LEVEL,
    EVENT_CHANNEL,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_PRESET,
    LOGGER,
)
from .dynet import DynetPacket, PacketError
from .event import DynetEvent
from .inbound import DynetInbound
from .opcodes import SyncType


class Dynalite:
    """Class to represent the interaction with Dynalite."""

    def __init__(self, broadcast_func):
        """Initialize the class."""
        self.host = None
        self.port = None
        self.loop = None
        self.broadcast_func = broadcast_func
        self._in_buffer = []
        self._out_buffer = []
        self._last_sent = None
        self._message_delay = 200
        self._sending = False
        self._reader = None
        self._writer = None
        self.resetting = False

    async def connect_internal(self):
        """Create the actual connection to Dynet."""
        try:
            self._reader, self._writer = await asyncio.open_connection(
                self.host, self.port
            )
            return True
        except (ValueError, OSError, asyncio.TimeoutError) as err:
            LOGGER.warning("Could not connect to Dynet (%s)", err)
            return False

    async def connect(self, host, port):
        """Connect to Dynet."""
        self.host = host
        self.port = port
        LOGGER.debug("Connecting to Dynet on %s:%s", host, port)
        if not self.loop:
            self.loop = asyncio.get_running_loop()
        self.resetting = False
        result = await self.connect_internal()
        if result and not self.resetting:
            self.loop.create_task(self.reader_loop())
        return result

    async def reader_loop(self):
        """Loop to read from the stream reader and reconnect if necessary."""
        while True:
            self.write()  # write if there is something in the buffers
            try:
                data = await self._reader.read(100)
                if len(data) > 0:
                    self.receive(data)
                    continue
            except ConnectionResetError:
                pass
            # we got disconnected or EOF
            if self.resetting:
                self._reader = None
                return  # stop loop
            self._reader = None
            self._writer = None
            self.broadcast(DynetEvent(event_type=EVENT_DISCONNECTED, data={}))
            await asyncio.sleep(1)  # Don't overload the network
            while not await self.connect_internal():
                if self.resetting:
                    self._reader = None
                    return  # stop loop
                await asyncio.sleep(1)  # Don't overload the network
            self.broadcast(DynetEvent(event_type=EVENT_CONNECTED, data={}))

    def broadcast(self, event):
        """Broadcast an event to all listeners - queue."""
        self.loop.call_soon(self.broadcast_func, event)

    def set_channel_level(self, area, channel, level, fade):
        """Set the level of a channel."""
        packet = DynetPacket.set_channel_level_packet(area, channel, level, fade)
        self.write(packet)
        broadcast_data = {
            CONF_AREA: area,
            CONF_CHANNEL: channel,
            CONF_TRGT_LEVEL: int(255 - 254.0 * level),
            CONF_ACTION: CONF_ACTION_CMD,
        }
        self.broadcast(DynetEvent(event_type=EVENT_CHANNEL, data=broadcast_data))

    def select_preset(self, area, preset, fade):
        """Select a preset in an area."""
        packet = DynetPacket.select_area_preset_packet(area, preset, fade)
        self.write(packet)
        broadcast_data = {
            CONF_AREA: area,
            CONF_PRESET: preset,
        }
        self.broadcast(DynetEvent(event_type=EVENT_PRESET, data=broadcast_data))

    def request_channel_level(self, area, channel):
        """Request a level for a specific channel."""
        packet = DynetPacket.request_channel_level_packet(area, channel)
        self.write(packet)

    def stop_channel_fade(self, area, channel):
        """Stop fading of a channel - async."""
        packet = DynetPacket.stop_channel_fade_packet(area, channel)
        self.write(packet)

    def request_area_preset(self, area):
        """Request current preset of an area."""
        packet = DynetPacket.request_area_preset_packet(area)
        self.write(packet)

    def receive(self, data=None):
        """Handle data that was received."""
        if data is not None:
            for byte in data:
                self._in_buffer.append(int(byte))
        if len(self._in_buffer) < 8:
            LOGGER.debug(
                "Received %d bytes, not enough to process: %s",
                len(self._in_buffer),
                self._in_buffer,
            )
        packet = None
        while len(self._in_buffer) >= 8 and packet is None:
            first_byte = self._in_buffer[0]
            if SyncType.has_value(first_byte):
                if first_byte == SyncType.DEBUG_MSG.value:
                    bytemsg = "".join(chr(c) for c in self._in_buffer[1:7])
                    LOGGER.debug("Dynet DEBUG message %s", bytemsg)
                    self._in_buffer = self._in_buffer[8:]
                    continue
                if first_byte == SyncType.DEVICE.value:
                    LOGGER.debug(
                        "Not handling Dynet DEVICE message %s", self._in_buffer[:8]
                    )
                    self._in_buffer = self._in_buffer[8:]
                    continue
                if first_byte == SyncType.LOGICAL.value:
                    try:
                        packet = DynetPacket(msg=self._in_buffer[:8])
                    except PacketError as err:
                        LOGGER.warning(err)
                        packet = None
                else:
                    LOGGER.error("wrong first_byte - we shouldn't get here")
            if packet is None:
                hex_string = ":".join("{:02x}".format(c) for c in self._in_buffer[:8])
                LOGGER.debug(
                    "Unable to process packet %s - moving one byte forward", hex_string
                )
                del self._in_buffer[0]
                continue
            self._in_buffer = self._in_buffer[8:]
            LOGGER.debug("Have packet: %s", packet)
            if hasattr(packet, "opcode_type") and packet.opcode_type is not None:
                inbound_handler = DynetInbound()
                if hasattr(inbound_handler, packet.opcode_type.lower()):
                    event = getattr(inbound_handler, packet.opcode_type.lower())(packet)
                    if event:
                        self.broadcast(event)
                else:
                    LOGGER.debug(
                        "Unhandled Dynet Inbound (%s): %s", packet.opcode_type, packet
                    )
            else:
                LOGGER.debug("Unhandled Dynet Inbound: %s", packet)
        # If there is still buffer to process - start again
        if len(self._in_buffer) >= 8:
            self.loop.call_soon(self.receive)

    def write(self, new_packet=None):
        """Write a packet or trigger write loop."""
        if new_packet is not None:
            self._out_buffer.append(new_packet)
        if self._writer is None:
            LOGGER.debug("write before transport is ready. queuing")
            return
        if self._sending:
            LOGGER.debug("Connection busy - queuing packet")
            self.loop.call_later(1, self.write)
            return
        if self._last_sent is None:
            self._last_sent = int(round(time.time() * 1000))
        current_milli_time = int(round(time.time() * 1000))
        elapsed = current_milli_time - self._last_sent
        delay = 0 - (elapsed - self._message_delay)
        if delay > 0:
            self.loop.call_later(delay / 1000, self.write)
            return
        if len(self._out_buffer) == 0:
            return
        packet = self._out_buffer[0]
        self._sending = True
        msg = bytearray()
        msg.append(packet.sync)
        msg.append(packet.area)
        msg.append(packet.data[0])
        msg.append(packet.command)
        msg.append(packet.data[1])
        msg.append(packet.data[2])
        msg.append(packet.join)
        msg.append(packet.chk)
        self._writer.write(msg)
        LOGGER.debug("Dynet Sent: %s", msg)
        self._last_sent = int(round(time.time() * 1000))
        self._sending = False
        del self._out_buffer[0]
        if len(self._out_buffer) > 0:
            self.loop.call_later(self._message_delay / 1000, self.write)

    async def async_reset(self):
        """Close sockets and timers."""
        self.resetting = True
        # Wait for reader to also close
        while self._reader:
            if self._writer:
                temp_writer = self._writer
                self._writer = None
                temp_writer.close()
                await temp_writer.wait_closed()
            await asyncio.sleep(1)



"""Class to represent an event on the Dynet network."""
import json


class DynetEvent:
    """Class to represent an event on the Dynet network."""

    def __init__(self, event_type=None, message=None, data=None):
        """Initialize the event."""
        self.event_type = event_type.upper() if event_type else None
        self.msg = message
        self.data = data

    def __repr__(self):
        """Print the event."""
        return json.dumps(self.__dict__)



"""Class to create devices from a Dynalite hub."""

import asyncio

from .const import (
    CONF_ACT_LEVEL,
    CONF_ACTION,
    CONF_ACTION_CMD,
    CONF_ACTION_REPORT,
    CONF_ACTION_STOP,
    CONF_ACTIVE,
    CONF_ACTIVE_INIT,
    CONF_ACTIVE_OFF,
    CONF_ACTIVE_ON,
    CONF_ALL,
    CONF_AREA,
    CONF_AREA_OVERRIDE,
    CONF_AUTO_DISCOVER,
    CONF_CHANNEL,
    CONF_CHANNEL_CLASS,
    CONF_CHANNEL_COVER,
    CONF_CHANNEL_TYPE,
    CONF_CLOSE_PRESET,
    CONF_DEFAULT,
    CONF_DEVICE_CLASS,
    CONF_DURATION,
    CONF_FADE,
    CONF_HIDDEN_ENTITY,
    CONF_HOST,
    CONF_NAME,
    CONF_NO_DEFAULT,
    CONF_NONE,
    CONF_OPEN_PRESET,
    CONF_POLL_TIMER,
    CONF_PORT,
    CONF_PRESET,
    CONF_ROOM,
    CONF_ROOM_OFF,
    CONF_ROOM_ON,
    CONF_STOP_PRESET,
    CONF_TEMPLATE,
    CONF_TILT_TIME,
    CONF_TIME_COVER,
    CONF_TRGT_LEVEL,
    CONF_TRIGGER,
    DEFAULT_CHANNEL_TYPE,
    DEFAULT_NAME,
    DEFAULT_PORT,
    DEFAULT_PRESETS,
    DEFAULT_TEMPLATES,
    EVENT_CHANNEL,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_PRESET,
    LOGGER,
)
from .cover import DynaliteTimeCoverDevice, DynaliteTimeCoverWithTiltDevice
from .dynalite import Dynalite
from .dynalitebase import DynaliteBaseDevice
from .light import DynaliteChannelLightDevice
from .switch import (
    DynaliteChannelSwitchDevice,
    DynaliteDualPresetSwitchDevice,
    DynalitePresetSwitchDevice,
)


class BridgeError(Exception):
    """For errors in the Dynalite bridge."""

    def __init__(self, message):
        """Initialize the exception."""
        self.message = message
        super().__init__(message)


class DynaliteDevices:
    """Manages a single Dynalite bridge."""

    def __init__(self, loop=None, new_device_func=None, update_device_func=None):
        """Initialize the system."""
        self.host = None
        self.port = None
        self.name = None
        self.poll_timer = None
        self.default_fade = None
        self.active = None
        self.auto_discover = None
        self.loop = loop
        self.new_device_func = new_device_func
        self.update_device_func = update_device_func
        self.configured = False
        self.connected = False
        self.added_presets = {}
        self.added_channels = {}
        self.added_room_switches = {}
        self.added_time_covers = {}
        self.waiting_devices = []
        self.timer_active = False
        self.timer_callbacks = set()
        self.template = {}
        self.area = {}
        self._dynalite = Dynalite(broadcast_func=self.handle_event)
        self.resetting = False

    async def async_setup(self):
        """Set up a Dynalite bridge based on host parameter in the config."""
        LOGGER.debug("bridge async_setup")
        if not self.loop:
            self.loop = asyncio.get_running_loop()
        # Run the dynalite object. Assumes self.configure() has been called
        self.resetting = False
        self.connected = await self._dynalite.connect(self.host, self.port)
        return self.connected

    def configure(self, config):
        """Configure a Dynalite bridge based on host parameter in the config."""
        LOGGER.debug("bridge async_configure - %s", config)
        self.configured = False
        # insert the global values
        self.host = config[CONF_HOST]
        self.port = config.get(CONF_PORT, DEFAULT_PORT)
        self.name = config.get(CONF_NAME, f"{DEFAULT_NAME}-{self.host}")
        self.auto_discover = config.get(CONF_AUTO_DISCOVER, False)
        self.active = config.get(CONF_ACTIVE, CONF_ACTIVE_INIT)
        if self.active is True:
            self.active = CONF_ACTIVE_ON
        if self.active is False:
            self.active = CONF_ACTIVE_OFF
        self.poll_timer = config.get(CONF_POLL_TIMER, 1.0)
        self.default_fade = config.get(CONF_DEFAULT, {}).get(CONF_FADE, 0)
        # create the templates
        config_templates = config.get(CONF_TEMPLATE, {})
        for template in DEFAULT_TEMPLATES:
            self.template[template] = {}
            cur_template = config_templates.get(template, {})
            for conf in DEFAULT_TEMPLATES[template]:
                self.template[template][conf] = cur_template.get(
                    conf, DEFAULT_TEMPLATES[template][conf]
                )
        # create default presets
        config_presets = config.get(CONF_PRESET, DEFAULT_PRESETS)
        default_presets = {}
        for preset in config_presets:
            cur_config = config_presets[preset]
            default_presets[int(preset)] = {
                CONF_NAME: cur_config.get(CONF_NAME, f"Preset {preset}"),
                CONF_FADE: cur_config.get(CONF_FADE, self.default_fade),
            }
        # create the areas with their channels and presets
        for area_val in config.get(CONF_AREA, {}):  # may be a string '123'
            area = int(area_val)
            area_config = config[CONF_AREA].get(area_val)
            self.area[area] = {
                CONF_NAME: area_config.get(CONF_NAME, f"Area {area}"),
                CONF_FADE: area_config.get(CONF_FADE, self.default_fade),
            }
            if CONF_TEMPLATE in area_config:
                self.area[area][CONF_TEMPLATE] = area_config[CONF_TEMPLATE]
            area_presets = {}
            area_channels = {}
            # User defined presets and channels first, then template presets, then defaults
            for preset in area_config.get(CONF_PRESET, {}):
                preset_config = area_config[CONF_PRESET][preset]
                area_presets[int(preset)] = {
                    CONF_NAME: preset_config.get(CONF_NAME, f"Preset {preset}"),
                    CONF_FADE: preset_config.get(CONF_FADE, self.area[area][CONF_FADE]),
                }
            for channel in area_config.get(CONF_CHANNEL, {}):
                channel_config = area_config[CONF_CHANNEL][channel]
                area_channels[int(channel)] = {
                    CONF_NAME: channel_config.get(CONF_NAME, f"Channel {channel}"),
                    CONF_FADE: channel_config.get(
                        CONF_FADE, self.area[area][CONF_FADE]
                    ),
                }
            # add the entities implicitly defined by templates
            template = area_config.get(CONF_TEMPLATE)
            if template:
                # Which type of value is a specific CONF
                conf_presets = [
                    CONF_ROOM_ON,
                    CONF_ROOM_OFF,
                    CONF_TRIGGER,
                    CONF_OPEN_PRESET,
                    CONF_CLOSE_PRESET,
                    CONF_STOP_PRESET,
                ]
                conf_values = [CONF_CHANNEL_CLASS, CONF_DURATION, CONF_TILT_TIME]
                conf_channels = [CONF_CHANNEL_COVER]

                for conf in self.template[template]:
                    conf_value = area_config.get(conf, self.template[template][conf])
                    if conf in conf_presets:
                        preset = int(conf_value)
                        if preset not in area_presets:
                            area_presets[preset] = {
                                CONF_NAME: f"Preset {preset}",
                                CONF_FADE: self.area[area][CONF_FADE],
                                # Trigger is the only exception
                                CONF_HIDDEN_ENTITY: (template != CONF_TRIGGER),
                            }
                        self.area[area][conf] = preset
                    elif conf in conf_channels:
                        channel = int(conf_value)
                        if channel not in area_channels:
                            area_channels[channel] = {
                                CONF_NAME: f"Channel {channel}",
                                CONF_FADE: self.area[area][CONF_FADE],
                                CONF_HIDDEN_ENTITY: True,
                            }
                        self.area[area][conf] = channel
                    else:
                        assert conf in conf_values
                        self.area[area][conf] = conf_value
            # Default presets
            if not area_config.get(CONF_NO_DEFAULT, False):
                for preset in default_presets:
                    if preset not in area_presets:
                        area_presets[preset] = default_presets[preset]
            self.area[area][CONF_PRESET] = area_presets
            self.area[area][CONF_CHANNEL] = area_channels
            # now register the channels and presets and ask for initial status if needed
            if self.active in [CONF_ACTIVE_INIT, CONF_ACTIVE_ON]:
                self._dynalite.request_area_preset(area)
            for channel in area_channels:
                self.create_channel_if_new(area, channel)
                if self.active in [CONF_ACTIVE_INIT, CONF_ACTIVE_ON]:
                    self._dynalite.request_channel_level(area, channel)
            for preset in area_presets:
                self.create_preset_if_new(area, preset)

        # register the rooms (switches on presets 1/4)
        # all the devices should be created for channels and presets
        self.register_rooms()
        # register the time covers
        self.register_time_covers()
        # callback for all devices
        if self.new_device_func and self.waiting_devices:
            self.new_device_func(self.waiting_devices)
            self.waiting_devices = []
        self.configured = True

    def register_rooms(self):
        """Register the room switches from two normal presets each."""
        for area, area_config in self.area.items():
            if area_config.get(CONF_TEMPLATE, "") == CONF_ROOM:
                if area in self.added_room_switches:
                    continue
                new_device = DynaliteDualPresetSwitchDevice(area, self)
                self.added_room_switches[area] = new_device
                new_device.set_device(
                    1, self.added_presets[area][area_config[CONF_ROOM_ON]]
                )
                new_device.set_device(
                    2, self.added_presets[area][area_config[CONF_ROOM_OFF]]
                )
                self.register_new_device("switch", new_device, False)

    def register_time_covers(self):
        """Register the time covers from three presets and a channel each."""
        for area, area_config in self.area.items():
            if area_config.get(CONF_TEMPLATE, "") == CONF_TIME_COVER:
                if area in self.added_time_covers:
                    continue
                if area_config[CONF_TILT_TIME] == 0:
                    new_device = DynaliteTimeCoverDevice(area, self)
                else:
                    new_device = DynaliteTimeCoverWithTiltDevice(area, self)
                self.added_time_covers[area] = new_device
                new_device.set_device(
                    1, self.added_presets[area][area_config[CONF_OPEN_PRESET]]
                )
                new_device.set_device(
                    2, self.added_presets[area][area_config[CONF_CLOSE_PRESET]]
                )
                new_device.set_device(
                    3, self.added_presets[area][area_config[CONF_STOP_PRESET]]
                )
                if area_config[CONF_CHANNEL_COVER] != 0:
                    channel_device = self.added_channels[area][
                        area_config[CONF_CHANNEL_COVER]
                    ]
                else:
                    channel_device = DynaliteBaseDevice(area, self)
                new_device.set_device(4, channel_device)
                self.register_new_device("cover", new_device, False)

    def register_new_device(self, category, device, hidden):
        """Register a new device and group all the ones prior to CONFIGURED event together."""
        # after initial configuration, every new device gets sent on its own. The initial ones are bunched together
        if not hidden:
            if self.configured:
                if self.new_device_func:
                    self.new_device_func([device])
            else:  # send all the devices together when configured
                self.waiting_devices.append(device)

    @property
    def available(self):
        """Return whether bridge is available."""
        return self.connected

    def update_device(self, device):
        """Update one or more devices."""
        if self.update_device_func:
            self.update_device_func(device)

    def handle_event(self, event=None):
        """Handle all events."""
        LOGGER.debug("handle_event - type=%s event=%s", event.event_type, event.data)
        if event.event_type == EVENT_CONNECTED:
            LOGGER.debug("Received CONNECTED message")
            self.connected = True
            self.update_device(CONF_ALL)
        elif event.event_type == EVENT_DISCONNECTED:
            LOGGER.debug("Received DISCONNECTED message")
            self.connected = False
            self.update_device(CONF_ALL)
        elif event.event_type == EVENT_PRESET:
            LOGGER.debug("Received PRESET message")
            self.handle_preset_selection(event)
        elif event.event_type == EVENT_CHANNEL:
            LOGGER.debug("Received PRESET message")
            self.handle_channel_change(event)
        else:
            LOGGER.debug(
                "Received unknown message type=%s data=%s", event.event_type, event.data
            )

    def ensure_area(self, area):
        """Configure a default area if it is not yet in config."""
        if area not in self.area:
            LOGGER.debug("adding area %s that is not in config", area)
            self.area[area] = {CONF_NAME: f"Area {area}", CONF_FADE: self.default_fade}

    def create_preset_if_new(self, area, preset):
        """Register a new preset."""
        LOGGER.debug("create_preset_if_new - area=%s preset=%s", area, preset)
        # if already configured, ignore
        if self.added_presets.get(area, {}).get(preset, False):
            return
        # if no autodiscover and not in config, ignore
        if not self.auto_discover:
            if not self.area.get(area, {}).get(CONF_PRESET, {}).get(preset, False):
                raise BridgeError(
                    f"No auto discovery and unknown preset (area {area} preset {preset}"
                )
        self.ensure_area(area)
        area_config = self.area[area]
        if CONF_PRESET not in area_config:
            area_config[CONF_PRESET] = {}
        if preset not in area_config[CONF_PRESET]:
            area_config[CONF_PRESET][preset] = {
                CONF_NAME: f"Preset {preset}",
                CONF_FADE: area_config[CONF_FADE],
            }
            # if the area is a template is a template, new presets should be hidden
            if area_config.get(CONF_TEMPLATE, False):
                area_config[CONF_PRESET][preset][CONF_HIDDEN_ENTITY] = True
        hidden = area_config[CONF_PRESET][preset].get(CONF_HIDDEN_ENTITY, False)
        new_device = DynalitePresetSwitchDevice(area, preset, self,)
        new_device.set_level(0)
        self.register_new_device("switch", new_device, hidden)
        if area not in self.added_presets:
            self.added_presets[area] = {}
        self.added_presets[area][preset] = new_device
        LOGGER.debug(
            "Creating Dynalite preset area=%s preset=%s hidden=%s", area, preset, hidden
        )

    def handle_preset_selection(self, event=None):
        """Change the selected preset."""
        LOGGER.debug("handle_preset_selection - event=%s", event.data)
        area = event.data[CONF_AREA]
        preset = event.data[CONF_PRESET]
        try:
            self.create_preset_if_new(area, preset)
        except BridgeError:
            # Unknown and no autodiscover
            return
        # Update all the preset devices
        for cur_preset_in_area in self.added_presets[area]:
            device = self.added_presets[area][cur_preset_in_area]
            if cur_preset_in_area == preset:
                device.set_level(1)
            else:
                device.set_level(0)
            self.update_device(device)
        # If active is set to full, query all channels in the area
        if self.active == CONF_ACTIVE_ON:
            for channel in self.area[area].get(CONF_CHANNEL, {}):
                self._dynalite.request_channel_level(area, channel)

    def create_channel_if_new(self, area, channel):
        """Register a new channel."""
        LOGGER.debug("create_channel_if_new - area=%s, channel=%s", area, channel)
        if channel == CONF_ALL:
            return
        # if already configured, ignore
        if self.added_channels.get(area, {}).get(channel, False):
            return
        # if no autodiscover and not in config, ignore
        if not self.auto_discover:
            if not self.area.get(area, {}).get(CONF_CHANNEL, {}).get(channel, False):
                raise BridgeError(
                    f"No auto discovery and unknown channel (area {area} channel {channel}"
                )
        self.ensure_area(area)
        area_config = self.area[area]
        if CONF_CHANNEL not in area_config:
            area_config[CONF_CHANNEL] = {}
        if channel not in area_config[CONF_CHANNEL]:
            area_config[CONF_CHANNEL][channel] = {
                CONF_NAME: f"Channel {channel}",
                CONF_FADE: area_config[CONF_FADE],
            }
            # if the area is a template is a template, new channels should be hidden
            if area_config.get(CONF_TEMPLATE, False):
                area_config[CONF_CHANNEL][channel][CONF_HIDDEN_ENTITY] = True
        channel_config = area_config[CONF_CHANNEL][channel]
        LOGGER.debug("create_channel_if_new - channel_config=%s", channel_config)
        channel_type = channel_config.get(
            CONF_CHANNEL_TYPE, DEFAULT_CHANNEL_TYPE
        ).lower()
        hidden = channel_config.get(CONF_HIDDEN_ENTITY, False)
        if channel_type == "light":
            new_device = DynaliteChannelLightDevice(area, channel, self,)
            self.register_new_device("light", new_device, hidden)
        elif channel_type == "switch":
            new_device = DynaliteChannelSwitchDevice(area, channel, self,)
            self.register_new_device("switch", new_device, hidden)
        else:
            LOGGER.info("unknown chnanel type %s - ignoring", channel_type)
            return
        if area not in self.added_channels:
            self.added_channels[area] = {}
        self.added_channels[area][channel] = new_device
        LOGGER.debug("Creating Dynalite channel area=%s channel=%s", area, channel)

    def handle_channel_change(self, event=None):
        """Change the level of a channel."""
        LOGGER.debug("handle_channel_change - event=%s", event.data)
        LOGGER.debug("handle_channel_change called event = %s", event.msg)
        area = event.data[CONF_AREA]
        channel = event.data[CONF_CHANNEL]
        try:
            self.create_channel_if_new(area, channel)
        except BridgeError:
            # Unknown and no autodiscover
            return
        action = event.data[CONF_ACTION]
        if action == CONF_ACTION_REPORT:
            actual_level = (255 - event.data[CONF_ACT_LEVEL]) / 254
            target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254
            channel_to_set = self.added_channels[area][channel]
            channel_to_set.update_level(actual_level, target_level)
            self.update_device(channel_to_set)
        elif action == CONF_ACTION_CMD:
            target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254
            # when there is only a "set channel level" command, assume that this is both the actual and the target
            actual_level = target_level
            channel_to_set = self.added_channels[area][channel]
            channel_to_set.update_level(actual_level, target_level)
            self.update_device(channel_to_set)
        elif action == CONF_ACTION_STOP:
            if channel == CONF_ALL:
                for channel in self.added_channels.get(area, {}):
                    channel_to_set = self.added_channels[area][channel]
                    channel_to_set.stop_fade()
                    self.update_device(channel_to_set)
            else:
                channel_to_set = self.added_channels[area][channel]
                channel_to_set.stop_fade()
                self.update_device(channel_to_set)
        else:
            LOGGER.error("unknown action for channel change %s", action)

    def add_timer_listener(self, callback_func):
        """Add a listener to the timer and start if needed."""
        self.timer_callbacks.add(callback_func)
        if not self.timer_active:
            self.loop.call_later(1, self.timer_func)
            self.timer_active = True

    def remove_timer_listener(self, callback_func):
        """Remove a listener from a timer."""
        self.timer_callbacks.discard(callback_func)

    def timer_func(self):
        """Call callbacks and either schedule timer or stop."""
        if self.timer_callbacks and not self.resetting:
            for callback in self.timer_callbacks:
                self.loop.call_soon(callback)
            self.loop.call_later(1, self.timer_func)
        else:
            self.timer_active = False

    def set_channel_level(self, area, channel, level, fade):
        """Set the level for a channel."""
        fade = self.area[area][CONF_CHANNEL][channel][CONF_FADE]
        self._dynalite.set_channel_level(area, channel, level, fade)

    def select_preset(self, area, preset):
        """Select a preset in an area."""
        fade = self.area[area][CONF_PRESET][preset][CONF_FADE]
        self._dynalite.select_preset(area, preset, fade)

    def get_channel_name(self, area, channel):
        """Return the name of a channel."""
        return f"{self.area[area][CONF_NAME]} {self.area[area][CONF_CHANNEL][channel][CONF_NAME]}"

    def get_channel_fade(self, area, channel):
        """Return the fade of a channel."""
        return self.area[area][CONF_CHANNEL][channel][CONF_FADE]

    def get_preset_name(self, area, preset):
        """Return the name of a preset."""
        return f"{self.area[area][CONF_NAME]} {self.area[area][CONF_PRESET][preset][CONF_NAME]}"

    def get_preset_fade(self, area, preset):
        """Return the fade of a preset."""
        return self.area[area][CONF_PRESET][preset][CONF_FADE]

    def get_multi_name(self, area):
        """Return the name of a multi-device."""
        return self.area[area][CONF_NAME]

    def get_device_class(self, area):
        """Return the class for a blind."""
        return self.area[area][CONF_DEVICE_CLASS]

    def get_master_area(self, area):
        """Get the master area when combining entities from different Dynet areas to the same area."""
        if area not in self.area:
            LOGGER.error("get_master_area - we should not get here")
            raise BridgeError(f"get_master_area - area {area} is not in config")
        area_config = self.area[area]
        master_area = area_config[CONF_NAME]
        if CONF_AREA_OVERRIDE in area_config:
            override_area = area_config[CONF_AREA_OVERRIDE]
            master_area = override_area if override_area.lower() != CONF_NONE else ""
        return master_area

    async def async_reset(self):
        """Reset the connections and timers."""
        self.resetting = True
        await self._dynalite.async_reset()
        while self.timer_active:
            await asyncio.sleep(1)



"""
Class that handles inbound requests on the Dynet network and fires events.

@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

from .const import (
    CONF_ACT_LEVEL,
    CONF_ACTION,
    CONF_ACTION_CMD,
    CONF_ACTION_REPORT,
    CONF_ACTION_STOP,
    CONF_ALL,
    CONF_AREA,
    CONF_CHANNEL,
    CONF_FADE,
    CONF_JOIN,
    CONF_PRESET,
    CONF_TRGT_LEVEL,
    EVENT_CHANNEL,
    EVENT_PRESET,
)
from .event import DynetEvent


class DynetInbound:
    """Class to handle inboud Dynet packets."""

    def __init__(self):
        """Initialize the object."""
        self._logger = None

    @staticmethod
    def preset(packet):
        """Handle a preset that was selected."""
        if packet.command > 3:
            packet.preset = packet.command - 6
        else:
            packet.preset = packet.command
        packet.preset = (packet.preset + (packet.data[2] * 8)) + 1
        packet.fade = (packet.data[0] + (packet.data[1] * 256)) * 0.02
        return DynetEvent(
            event_type=EVENT_PRESET,
            message=(
                "Area %d Preset %d Fade %d seconds."
                % (packet.area, packet.preset, packet.fade)
            ),
            data={
                CONF_AREA: packet.area,
                CONF_PRESET: packet.preset,
                CONF_FADE: packet.fade,
                CONF_JOIN: packet.join,
            },
        )

    def preset_1(self, packet):
        """Handle preset 1 in banks of 8."""
        return self.preset(packet)

    def preset_2(self, packet):
        """Handle preset 2 in banks of 8."""
        return self.preset(packet)

    def preset_3(self, packet):
        """Handle preset 3 in banks of 8."""
        return self.preset(packet)

    def preset_4(self, packet):
        """Handle preset 4 in banks of 8."""
        return self.preset(packet)

    def preset_5(self, packet):
        """Handle preset 5 in banks of 8."""
        return self.preset(packet)

    def preset_6(self, packet):
        """Handle preset 6 in banks of 8."""
        return self.preset(packet)

    def preset_7(self, packet):
        """Handle preset 7 in banks of 8."""
        return self.preset(packet)

    def preset_8(self, packet):
        """Handle preset 8 in banks of 8."""
        return self.preset(packet)

    @staticmethod
    def report_preset(packet):
        """Report the current preset of an area."""
        packet.preset = packet.data[0] + 1
        return DynetEvent(
            event_type=EVENT_PRESET,
            message=("Current Area %d Preset is %d" % (packet.area, packet.preset)),
            data={
                CONF_AREA: packet.area,
                CONF_PRESET: packet.preset,
                CONF_JOIN: packet.join,
            },
        )

    @staticmethod
    def linear_preset(packet):
        """Report that preset was selected with fade."""
        packet.preset = packet.data[0] + 1
        packet.fade = (packet.data[1] + (packet.data[2] * 256)) * 0.02
        return DynetEvent(
            event_type=EVENT_PRESET,
            message=(
                "Area %d Preset %d Fade %d seconds."
                % (packet.area, packet.preset, packet.fade)
            ),
            data={
                CONF_AREA: packet.area,
                CONF_PRESET: packet.preset,
                CONF_FADE: packet.fade,
                CONF_JOIN: packet.join,
            },
        )

    @staticmethod
    def report_channel_level(packet):
        """Report the new level of a channel."""
        channel = packet.data[0] + 1
        target_level = packet.data[1]
        actual_level = packet.data[2]
        return DynetEvent(
            event_type=EVENT_CHANNEL,
            message=(
                "Area %d Channel %d Target Level %d Actual Level %d."
                % (packet.area, channel, target_level, actual_level)
            ),
            data={
                CONF_AREA: packet.area,
                CONF_CHANNEL: channel,
                CONF_ACTION: CONF_ACTION_REPORT,
                CONF_TRGT_LEVEL: target_level,
                CONF_ACT_LEVEL: actual_level,
                CONF_JOIN: packet.join,
            },
        )

    @staticmethod
    def set_channel_x_to_level_with_fade(packet, channel_offset):
        """Report that a channel was set to a specific level."""
        channel = ((packet.data[1] + 1) % 256) * 4 + channel_offset
        target_level = packet.data[0]
        return DynetEvent(
            event_type=EVENT_CHANNEL,
            message=(
                "Area %d Channel %d Target Level %d"
                % (packet.area, channel, target_level)
            ),
            data={
                CONF_AREA: packet.area,
                CONF_CHANNEL: channel,
                CONF_ACTION: CONF_ACTION_CMD,
                CONF_TRGT_LEVEL: target_level,
                CONF_JOIN: packet.join,
            },
        )

    def set_channel_1_to_level_with_fade(self, packet):
        """Report that channel 1 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 1)

    def set_channel_2_to_level_with_fade(self, packet):
        """Report that channel 2 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 2)

    def set_channel_3_to_level_with_fade(self, packet):
        """Report that channel 3 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 3)

    def set_channel_4_to_level_with_fade(self, packet):
        """Report that channel 4 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 4)

    @staticmethod
    def request_channel_level(packet):
        """Do nothing."""
        pass

    @staticmethod
    def stop_fading(packet):
        """Report that fading stopped for a channel or area."""
        channel = packet.data[0] + 1
        if channel == 256:  # all channels in area
            channel = CONF_ALL
        return DynetEvent(
            event_type=EVENT_CHANNEL,
            message=("Area %d Channel %s" % (packet.area, channel)),
            data={
                CONF_AREA: packet.area,
                CONF_CHANNEL: channel,
                CONF_ACTION: CONF_ACTION_STOP,
                CONF_JOIN: packet.join,
            },
        )

    # def fade_channel_area_to_preset(self, packet): XXX
    # """Report that a channel or area was set to a preset."""
    # channel = packet.data[0] + 1
    # packet.preset = packet.data[1] + 1
    # packet.fade = packet.data[2] * 0.02
    # if channel == 256:  # all channels in area
    # return DynetEvent(
    # event_type=EVENT_PRESET,
    # message=(
    # "Current Area %d Preset is %d fade %s"
    # % (packet.area, packet.preset, packet.fade)
    # ),
    # data={
    # CONF_AREA: packet.area,
    # CONF_PRESET: packet.preset,
    # CONF_FADE: packet.fade,
    # CONF_JOIN: packet.join,
    # },
    # )
    # else:
    # return DynetEvent(
    # event_type=EVENT_CHANNEL,
    # message=(
    # "Area %d Channel %s preset %s fade %s"
    # % (packet.area, channel, packet.preset, packet.fade)
    # ),
    # data={
    # CONF_AREA: packet.area,
    # CONF_CHANNEL: channel,
    # CONF_FADE: packet.fade,
    # CONF_ACTION: CONF_ACTION_CMD,
    # CONF_PRESET: packet.preset,
    # CONF_JOIN: packet.join,
    # },
    # )
